#include "stdafx.h"
#include "PlaylistManager.h"
#include "SysTrayPlay.h"
#include "MainHelper.h"
#include "MainWnd.h"
#include "PlaylistDlg.h"
#include "MP3InfoDlg.h"
#include "In/stream.h"
#include "In/dxhead.h"
#include "In/frame.h"
#include "In/MultiDllInput.h"
#include "Out/MultiDllOutput.h"
#include "WAEmuWnd.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

HANDLE hTagsThread = NULL;
HANDLE hAutomationThread = NULL;
extern CMainWnd *pMainFrame;
extern CPlaylistDlg *PlaylistDlg;
extern CMP3InfoDlg *MP3InfoDlg;
extern PlaylistManager m_PlaylistManager;
extern CMultiDllInput *stdIn;
extern CMultiDllOutput *stdOut;
extern CWAEmuWnd *WAEmuWnd;
extern CStringArray genres;

DWORD q;

bool decode_header_tag(unsigned long newhead, Frame *pfr) {
    if((newhead & 0xffe00000) != 0xffe00000)
        return false;
    if(!((newhead >> 17) & 3))
        return false;
    if(((newhead >> 12) & 0xf) == 0xf)
        return false;
    if(((newhead >> 10) & 0x3) == 0x3)
        return false;
    if((newhead & 0xffff0000) == 0xfffe0000)
        return false;

    if(newhead & (1 << 20)) {
        pfr->lsf = (newhead & (1 << 19)) ? 0x0 : 0x1;
        pfr->mpeg25 = 0;
    }
    else {
        pfr->lsf = 1;
        pfr->mpeg25 = 1;
    }

    pfr->lay = 4 - ((newhead >> 17) & 3);

    if(pfr->mpeg25)
        pfr->sampling_frequency = 6 + ((newhead >> 10) & 0x3);
    else
        pfr->sampling_frequency = ((newhead >> 10) & 0x3) + (pfr->lsf * 3);

    pfr->error_protection = ((newhead >> 16) & 0x1) ^ 0x1;
    pfr->bitrate_index = ((newhead >> 12) & 0xf);
    pfr->padding = ((newhead >> 9) & 0x1);
    pfr->extension = ((newhead >> 8) & 0x1);
    pfr->mode = ((newhead >> 6) & 0x3);
    pfr->mode_ext = ((newhead >> 4) & 0x3);
    pfr->copyright = ((newhead >> 3) & 0x1);
    pfr->original = ((newhead >> 2) & 0x1);
    pfr->emphasis = newhead & 0x3;

    pfr->stereo = (pfr->mode == MPG_MD_MONO) ? 1 : 2;

    if(!pfr->bitrate_index)
        return false;

    switch(pfr->lay) {
        case 1:
            pfr->jsbound = (pfr->mode == MPG_MD_JOINT_STEREO) ? (pfr->mode_ext << 2) + 4 : 32;
            pfr->framesize = GetBitrate(pfr->lsf, 1, pfr->bitrate_index) * 12000;
            pfr->framesize /= freqs[pfr->sampling_frequency];
            pfr->framesize = ((pfr->framesize + pfr->padding) << 2) - 4;
            break;
        case 2:
            pfr->jsbound = (pfr->mode == MPG_MD_JOINT_STEREO) ? (pfr->mode_ext << 2) + 4 : pfr->II_sblimit;
            pfr->framesize = GetBitrate(pfr->lsf, 2, pfr->bitrate_index) * 144000;
            pfr->framesize /= freqs[pfr->sampling_frequency];
            pfr->framesize += pfr->padding - 4;
            break;
        case 3:
            pfr->framesize = GetBitrate(pfr->lsf, 3, pfr->bitrate_index) * 144000;
            pfr->framesize /= freqs[pfr->sampling_frequency] << (pfr->lsf);
            pfr->framesize = pfr->framesize + pfr->padding - 4;
            break; 
        default:
            return false;
    }
    return true;
}



CachedTag::CachedTag() {
    duration = 0;
	genre = 0;
    read = false;
    artist = "";
    album = "";
    comment = "";
    playlist = "";
    title = "";
    year = "";
}

CachedTag::CachedTag(const CachedTag &copie) {
    read = copie.read;
    duration = copie.duration;
	genre = copie.genre;
    artist = copie.artist;
    album = copie.album;
    comment = copie.comment;
    playlist = copie.playlist;
    title = copie.title;
    year = copie.year;
}

CachedTag::CachedTag(CString nPlaylist) {
    read = false;
    duration = 0;
	genre = 0;
    artist = "";
    album = "";
    comment = "";
    playlist = nPlaylist;
    title = "";
    year = "";
}

CachedTag::~CachedTag() {
}

void CachedTag::copy(CachedTag copie) {
    read = copie.read;
    duration = copie.duration;
	genre = copie.genre;
    artist = copie.artist;
    album = copie.album;
    comment = copie.comment;
    playlist = copie.playlist;
    title = copie.title;
    year = copie.year;
}

CString CachedTag::GetAlbum() {
    if(!read)
        ReadTag();
    return album;
}

CString CachedTag::GetArtist() {
    if(!read)
        ReadTag();
    return artist;
}

CString CachedTag::GetComment() {
    if(!read)
        ReadTag();
    return comment;
}

int CachedTag::GetGenreI() {
    if(!read)
        ReadTag();
    return genre;
}


CString CachedTag::GetGenre() {
    if(!read)
        ReadTag();
	return genres[GetGenreI()];
}

CString CachedTag::GetFilename() {
    CString filename = playlist;
    int pos = filename.GetLength();
    return filename.Right(pos - filename.ReverseFind('\\') - 1);
}

CString CachedTag::GetDuration() {
    if(!read)
        ReadTag();
    CString dur = "(mm:ss)";
    if(duration != 0)
        dur.Format(_T(" (%02d:%02d)"), (int) (duration / 60), duration % 60);
    return dur;
}

int CachedTag::GetDurationI() {
    return duration;
}

CString CachedTag::GetPlaylist() {
    return playlist;
}

bool CachedTag::GetRead() 
{
    return read;
}

CString CachedTag::GetTag(int ShowingMode) {
    if(ShowingMode == 2)
        return GetFilename();
    else if(ShowingMode == 1) {
        if(GetArtist().GetLength() == 0)
            return GetTitle() + " " + GetDuration();
        else
            return GetArtist() + " - " + GetTitle() + " " + GetDuration();
    }
    else
        return GetTitle() + " " + GetDuration();
}

CString CachedTag::GetTitle() {
    if(!read)
        ReadTag();
    return title;
}

CString CachedTag::GetYear() {
    if(!read)
        ReadTag();
    return year;
}

void CachedTag::ReadTag(bool forceRead) {
    if(read && !forceRead)
        return;
	TCHAR title[5000];
	int length=0;
	stdIn->GetFileInfo((TCHAR *)(LPCTSTR)GetPlaylist(),(TCHAR *)(LPCTSTR)title,&length);
	SetTitle(title);
	SetDuration(length);

    StreamManager fileStreamManager;
    fileStreamManager.SetInName((TCHAR *)(LPCTSTR)playlist);

    bool Seek;
    unsigned long size = fileStreamManager.InitStream(&Seek);
    if(!size && fileStreamManager.GetInputMode() == imFile) {
        // Si ouverture chou
        return;
    }   

    int i;
    unsigned char hbuf[4096];
	memset(hbuf,0,4095);
    unsigned long newhead, oldhead;
    Frame frtmp;
    int riffsize = 0, id3v2size = 0, dur = 0;
    DWORD fLength = 0;
    DWORD numread;
    bool read_again;
    do {
        read_again = false;
        if(!fileStreamManager.ReadF(hbuf, 4, &numread))
            numread = 0;
        if(numread < 4)
            goto fault;

        newhead = ((unsigned long) hbuf[0] << 24)
                | ((unsigned long) hbuf[1] << 16)
                | ((unsigned long) hbuf[2] << 8)
                | (unsigned long) hbuf[3];

        if(!decode_header_tag(newhead, &frtmp)) {
            //Si le type n'est pas reconnu
            if(newhead == ('R' << 24) + ('I' << 16) + ('F' << 8) + 'F') {
                try {
                    fileStreamManager.SetFPointer(66, 0, FILE_CURRENT);
                    riffsize = 70;
                }
                catch(...) {
                    fileStreamManager.Close();
                    return;
                };
                read_again = true;
                continue;
            }
            if((newhead & (('I' << 24) + ('D' << 16) + ('3' << 8))) == (('I' << 24) + ('D' << 16) + ('3' << 8))) {
                try {
                    fileStreamManager.SetFPointer(2, 0, FILE_CURRENT);
                    fileStreamManager.ReadF(hbuf, 4, &numread);

                    id3v2size = (int) (hbuf[0] & 0x7f) * 0x200000
                              + (int) (hbuf[1] & 0x7f) * 0x4000
                              + (int) (hbuf[2] & 0x7f) * 0x80
                              + (int) (hbuf[3] & 0x7f);
                    fileStreamManager.SetFPointer(id3v2size, 0, FILE_CURRENT);
                }
                catch(...) {
                    fileStreamManager.Close();
                    return;
                }
                read_again = true;
                continue;
            }
            for(i = 0; i < 65536; i++) {
                memmove(&hbuf[0], &hbuf[1], 3);
                fileStreamManager.ReadF(hbuf + 3, 1, &numread);
                if(numread < 1 && fileStreamManager.GetInputMode() == imFile)
                    goto fault;
                oldhead = newhead >> 24;
                newhead <<= 8;
                newhead |= hbuf[3];
                newhead &= 0xffffffff;
                if(fileStreamManager.GetInputMode() == imHTTP && (newhead & 0xFFFFFFFF) == 0x0D0A0D0A)
                    break;
                if(decode_header_tag(newhead, &frtmp))
                    break;
            }
            if(i == 65536)
                goto fault;
        }
    }
    while(read_again);
    fLength = size;
    dur = (fLength - riffsize - id3v2size) * 8 / GetBitrate(frtmp.lsf, frtmp.lay, frtmp.bitrate_index) / 1000;

    try {
        int ssize;
        if(frtmp.lsf)
            ssize = (frtmp.stereo == 1) ? 9 : 17;
        else
            ssize = (frtmp.stereo == 1) ? 17 : 32;
        if(frtmp.error_protection)
            ssize += 2;
        if(fileStreamManager.GetInputMode() == imHTTP) {
            fileStreamManager.SetFPointer(0, 0, FILE_CURRENT);
            fileStreamManager.ReadF(hbuf, i + 5, &numread);
            CString tmp;
            CString tmp2 = "";
			unsigned char strToFin[10]="icy-name:";
			unsigned char *test = _mbsstr(hbuf,strToFin);
			if(test!=NULL) {
				int iName = (int)(test - hbuf) + 9;
				while(hbuf[iName] != 13) {
					tmp.Format(_T("%s%c"),tmp2,hbuf[iName]);
					tmp2=tmp;
					iName++;
				}
				SetTitle(tmp2);
				TRACE(_T("%s"), tmp);
			}
            return;
        }
        else {
            fileStreamManager.SetFPointer(ssize, 0, FILE_CURRENT);
            fileStreamManager.ReadF(hbuf, 4, &numread);
            if(numread == 4) {
                if(hbuf[0] == 'X' && hbuf[1] == 'i' && hbuf[2] == 'n' && hbuf[3] == 'g') {
                    char XingData[116];
                    fileStreamManager.ReadF((unsigned char *) XingData, 116, &numread);
                    if(numread == 116) {
                        int flags = ExtractI4((unsigned char*) XingData), frames;
                        if((flags & 1) > 0)
                            frames = ExtractI4((unsigned char *) XingData + 4);
                        else
                            frames = 0;
                        dur = (int) ((double) 1152 / (double) freqs[frtmp.sampling_frequency] * (double) frames);
                    }
                }
            }
        }
    }
    catch(...) {
        fileStreamManager.Close();
        return;
    }
    duration = dur;
    fault:

    if(size > 128 && fileStreamManager.GetInputMode() == imFile) {
        try {
            fileStreamManager.SetFPointer(-128, 0, FILE_END);

            char buf[31];
            fileStreamManager.ReadF((unsigned char *) buf, 3, &numread);
            buf[3] = '\0';
            if(strcmp(buf, "TAG")) {
                fileStreamManager.Close();
                return;
            }

            CString temp;
            //Title
            fileStreamManager.ReadF((unsigned char *) buf, 30, &numread);
            buf[30] = '\0';
            temp = buf;
            temp.TrimLeft();
            temp.TrimRight();
            SetTitle(temp);
            //Artist
            fileStreamManager.ReadF((unsigned char *) buf, 30, &numread);
            buf[30] = '\0';
            temp = buf;
            temp.TrimLeft();
            temp.TrimRight();
            SetArtist(temp);
            //Album
            fileStreamManager.ReadF((unsigned char *) buf, 30, &numread);
            buf[30] = '\0';
            temp = buf;
            temp.TrimLeft();
            temp.TrimRight();
            SetAlbum(temp);
            //Year
            fileStreamManager.ReadF((unsigned char *) buf, 4, &numread);
            buf[4] = '\0';
            temp = buf;
            temp.TrimLeft();
            temp.TrimRight();
            SetYear(temp);
            //Comment
            fileStreamManager.ReadF((unsigned char *) buf, 30, &numread);
            buf[30] = '\0';
            temp = buf;
            temp.TrimLeft();
            temp.TrimRight();
            SetComment(temp);
			//Genre
            fileStreamManager.ReadF((unsigned char *) buf, 1, &numread);
			SetGenreI((int)buf[0]);
        }
        catch(...) {
        };
    }
    else if(fileStreamManager.GetInputMode() == imHTTP) {
        fileStreamManager.SetFPointer(frtmp.framesize + 30, 0, FILE_BEGIN);
        char buf[500];
        int count = 0;
        CString temp;
        //Title
        fileStreamManager.ReadF((unsigned char *) buf, 1024, &numread);
        buf[numread] = '\0';
        temp = buf;
        temp.TrimLeft();
        temp.TrimRight();
        SetTitle(temp);
        read = false;
    }

    fileStreamManager.Close();
}
void CachedTag::SetAlbum(CString nAlbum) {
    read = true;
    album = nAlbum;
}

void CachedTag::SetArtist(CString nArtist) {
    read = true;
    artist = nArtist;
}

void CachedTag::SetComment(CString nComment) {
    read = true;
    comment = nComment;
}

void CachedTag::SetDuration(int nDuration) {
    read = true;
    duration = nDuration;
}

void CachedTag::SetGenreI(int nGenre) {
    read = true;
    genre = nGenre;
}

void CachedTag::SetPlaylist(CString nplaylist) {
    playlist = nplaylist;
}

void CachedTag::SetTitle(CString nTitle) {
    read = true;
    title = nTitle;
}

void CachedTag::SetYear(CString nYear) {
    read = true;
    year = nYear;
}


PlaylistManager::PlaylistManager() {
    maxSize = 0;
    taille = 0;
    CachedTags = (CachedTag * *) calloc(maxSize, sizeof(CachedTag *));
}

PlaylistManager::~PlaylistManager() {
    RemoveAll();
    delete CachedTags;
}

void PlaylistManager::Add(CString tag, int index) {
    CachedTag tmp(tag);
    if(index == -1)
        InsertAt(tmp, taille);
    else
        (*(CachedTags + index))->copy(tmp);
}

void PlaylistManager::Add(CachedTag tag, int index) {
    if(index == -1)
        InsertAt(tag, taille);
    else
        (*(CachedTags + index))->copy(tag);
}

void PlaylistManager::InsertAt(CachedTag tag, int index) {
    if(++taille == maxSize + 1) {
        maxSize = maxSize + PAS;
        CachedTags = (CachedTag * *) realloc(CachedTags, maxSize * sizeof(CachedTag *));
        for(int i = maxSize - PAS; i < maxSize; i++)
            *(CachedTags + i) = new CachedTag();
    }
    CachedTag *tmp = *(CachedTags + taille - 1);
    memmove(CachedTags + index + 1, CachedTags + index, (taille - index - 1) * sizeof(CachedTag *));
    *(CachedTags + index) = tmp;
    (*(CachedTags + index))->copy(tag);
}

CachedTag PlaylistManager::Get(int index) {
    CachedTag tmp;
    if(index < taille)
        return **(CachedTags + index);
    else
        return tmp;
}

int PlaylistManager::GetIndex(CString filename) {
    int posIndex;
    TRACE(_T("GetIndex %s %d\n"), filename, GetSize());
    if(GetSize() == 0)
        return -1;
    for(posIndex = 0; posIndex < GetSize() && Get(posIndex).GetPlaylist().Compare(filename); posIndex++)
        ;
    if(posIndex == GetSize())
        return -1;
    return posIndex;
}

CString PlaylistManager::GetAlbum(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetAlbum();
    else
        return "";
}

CString PlaylistManager::GetArtist(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetArtist();
    else
        return "";
}

CString PlaylistManager::GetComment(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetComment();
    else
        return "";
}

CString PlaylistManager::GetDuration(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetDuration();
    else
        return "";
}

int PlaylistManager::GetDurationI(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetDurationI();
    else
        return 0;
}

CString PlaylistManager::GetFilename(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetFilename();
    else
        return "";
}

CString PlaylistManager::GetGenre(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetGenre();
    else
        return "";
}

int PlaylistManager::GetGenreI(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetGenreI();
    else
        return 0;
}

CString PlaylistManager::GetPlaylist(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetPlaylist();
    else
        return "";
}

bool PlaylistManager::GetRead(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetRead();
    return false;
}
int PlaylistManager::GetShowingMode() {
    return ShowingMode;
}

CString PlaylistManager::GetTag(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetTag(ShowingMode);
    else
        return "";
}

CString PlaylistManager::GetTitle(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetTitle();
    else
        return "";
}

CString PlaylistManager::GetYear(int index) {
    if(index < taille)
        return (*(CachedTags + index))->GetYear();
    else
        return "";
}

int PlaylistManager::GetSize() {
    return taille;
}

void PlaylistManager::SetShowingMode(int nShowingMode) {
    ShowingMode = nShowingMode;
}

bool PlaylistManager::StopReadAll() {
    bool toret = (hTagsThread != NULL);
    if(toret) {
        ::TerminateThread(hTagsThread, 0);
        hTagsThread = NULL;
    }
    return toret;
}

bool PlaylistManager::StopPerformAutomation() {
    bool toret = (hAutomationThread != NULL);
    if(toret) {
        ::TerminateThread(hAutomationThread, 0);
        hAutomationThread = NULL;
    }
    return toret;
}

void ReadAllTagsThread() {
    for(UINT i = 0; i < (UINT) m_PlaylistManager.GetSize(); i++) {
        m_PlaylistManager.ReadTagAT(i, true);
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
            PlaylistDlg->ShowTimes();
            PlaylistDlg->m_playlistat.RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
        }
        Sleep(77);
    }
    if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
        PlaylistDlg->ShowTimes();
        PlaylistDlg->m_playlistat.RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
    }
    hTagsThread = NULL;
}

void PlaylistManager::ReadAllTags() {
    StopReadAll();
    hTagsThread = ::CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) ReadAllTagsThread, NULL, 0, &q);
    ::SetThreadPriority(hTagsThread, THREAD_PRIORITY_LOWEST);
}
void PerformAutomationThread() {
    bool b = false;
    if(pMainFrame->AutRemove || pMainFrame->AutRemovedup || pMainFrame->AutRescan || pMainFrame->AutSort)
        b = m_PlaylistManager.StopReadAll();
    if(pMainFrame->AutRescan)
        b = true;

    if(pMainFrame->AutRemove) {
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
            PlaylistDlg->m_playlistat.SetRedraw(FALSE);
        UINT i = 0;
        bool CurDeleted = false, deleted = false;
        while(i<(UINT) m_PlaylistManager.GetSize() && (UINT) m_PlaylistManager.GetSize()>0) {
            if(!PlaylistDlg->FileExists(m_PlaylistManager.GetPlaylist(i))) {
                deleted = true;
                if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
                    PlaylistDlg->m_playlistat.DeleteString(i);
                m_PlaylistManager.RemoveAt(i);
                if(i < pMainFrame->Current)
                    pMainFrame->Current--;
                else if(i == pMainFrame->Current && !CurDeleted)
                    CurDeleted = true;
            }
            else
                i++;
            Sleep(1);
        }
        if(CurDeleted) {
            pMainFrame->Current = 0;
            if(m_PlaylistManager.GetSize() > 0) {
                m_PlaylistManager.ReadTagAT(pMainFrame->Current);
                if(::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
                    ::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);
            }
            else {
                if(::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
                    ::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);
                CRect rect;
                if(MP3InfoDlg != NULL && MP3InfoDlg->m_hWnd != NULL)
                    MP3InfoDlg->SaveWindowPosAndDestroy();
            }
            pMainFrame->UpdMBTR();
        }
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL && deleted)
            PlaylistDlg->m_playlistat.SetSel(-1, FALSE);

        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
            PlaylistDlg->ShowFileName();
            PlaylistDlg->m_playlistat.SetRedraw(TRUE);
            PlaylistDlg->m_playlistat.Invalidate();
        }
    }

    if(pMainFrame->AutSort || pMainFrame->AutRemovedup) {
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
            PlaylistDlg->m_playlistat.SetRedraw(FALSE);

        pMainFrame->SortRec(0, m_PlaylistManager.GetSize() - 1, true);

        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
            PlaylistDlg->ShowFileName();
            PlaylistDlg->m_playlistat.SetTopIndex(PlaylistDlg->m_playlistat.GetFirstSelected()
                                               != -1
                                                ? PlaylistDlg->m_playlistat.GetFirstSelected()
                                                : pMainFrame->Current);
            PlaylistDlg->m_playlistat.SetRedraw(TRUE);
            PlaylistDlg->m_playlistat.Invalidate();
        }
    }

    if(pMainFrame->AutRemovedup) {
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
            PlaylistDlg->m_playlistat.SetRedraw(FALSE);
        UINT i = 0, j;
        bool deleted = false;
        while(i<(UINT) m_PlaylistManager.GetSize() && (UINT) m_PlaylistManager.GetSize()>0) {
            j = i + 1;
            while(j < (UINT) m_PlaylistManager.GetSize()) {
                if(m_PlaylistManager.GetPlaylist(i).Compare(m_PlaylistManager.GetPlaylist(j)) == 0) {
                    if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
                        PlaylistDlg->m_playlistat.DeleteString(j);
                    deleted = true;
                    m_PlaylistManager.RemoveAt(j);
                    if(j < pMainFrame->Current)
                        pMainFrame->Current--;
                    else if(j == pMainFrame->Current)
                        pMainFrame->Current = i;
                }
                else
                    j++;
                Sleep(1);
            }
            i++;
        }
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL && deleted)
            PlaylistDlg->m_playlistat.SetSel(-1, FALSE);

        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
            PlaylistDlg->ShowFileName();
            PlaylistDlg->m_playlistat.SetRedraw(TRUE);
            PlaylistDlg->m_playlistat.Invalidate();
        }
    }


    if(pMainFrame->AutRescan) {
        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
            PlaylistDlg->m_playlistat.SetRedraw(FALSE);
        UINT PLLPrev = m_PlaylistManager.GetSize();
        for(int i = 0; i < pMainFrame->AutDirs.GetSize(); i++)
            pMainFrame->FindMP3Files(pMainFrame->AutDirs[i], true);
        pMainFrame->SortRec(0, m_PlaylistManager.GetSize() - 1);

        if((pMainFrame->RandomFirst || pMainFrame->ShuffleMode) && PLLPrev == 0)
            pMainFrame->FirstInMyRandom = pMainFrame->Current = pMainFrame->MyRandom(true, true);
        else if(PLLPrev == 0) {
            pMainFrame->Current = 0;pMainFrame->FirstInMyRandom = 0xFFFFFFFF;
        }

        if(PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL) {
            if(m_PlaylistManager.GetSize() != PLLPrev) {
                PlaylistDlg->FillPlaylist();
                PlaylistDlg->ShowFileName();
            }
            PlaylistDlg->m_playlistat.SetRedraw(TRUE);
            PlaylistDlg->m_playlistat.Invalidate();
        }

        if(PLLPrev == 0) {
            if(::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
                ::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);

            bool first = true;
            int q = 0;
            do {
                if(!first) {
                    if(pMainFrame->Current < (UINT) m_PlaylistManager.GetSize() - 1)
                        pMainFrame->Current++;
                    else {
                        pMainFrame->Current = 0;
                        q++;
                        if(pMainFrame->m_Repeat != 1)
                            break;
                    }
                }
                else
                    first = false;
                break;
                Sleep(77);
            }
            while(q < 2);
            m_PlaylistManager.ReadTagAT(pMainFrame->Current);
            pMainFrame->UpdMBTR();
        }
    }

    if(b)
        m_PlaylistManager.ReadAllTags();
}

void PlaylistManager::PerformAutomation() {
    StopPerformAutomation();
    if(hAutomationThread == NULL) {
        DWORD q;
        hAutomationThread = ::CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) PerformAutomationThread, NULL, 0, &q);
        ::SetThreadPriority(hAutomationThread, THREAD_PRIORITY_LOWEST);
    }
}
void PlaylistManager::ReadTagAT(int index, bool forceRead) {
    if(index < taille)
        (*(CachedTags + index))->ReadTag(forceRead);
}
void PlaylistManager::RemoveAt(int index) {
    CachedTag *tmp = *(CachedTags + index);
    memmove(CachedTags + index, CachedTags + index + 1, (taille - index - 1) * sizeof(CachedTag *));
    *(CachedTags + taille - 1) = tmp;
    taille--;
    if(taille + PAS == maxSize) {
        for(int i = maxSize - PAS; i < maxSize; i++)
            delete * (CachedTags + i);
        maxSize = maxSize - PAS;
        CachedTags = (CachedTag * *) realloc(CachedTags, maxSize * sizeof(CachedTag *));
    }
}

void PlaylistManager::RemoveAll() {
    for(int i = taille - 1; i >= 0; i--)
        RemoveAt(i);
}

